handgunprojectileHandler = {
	shellNum = 1,
	shells = {},
	defaultShell = {active = false},
}

function init()
	RegisterTool("handgun", "Hand Gun", "MOD/vox/handgun.vox")
	SetBool("game.tool.handgun.enabled", true)
	SetFloat("game.tool.handgun.ammo", 101)

	damage = 1
	gravity = Vec(0, 0, 0)
	velocity = 1.6
	rocketvelocity = 0.8
	rocketgravity = Vec(0, -5, 0)
	shotDelay = 0.2
	ammo = 7
	mags = 6
	
	switchsound = GetBool("savegame.mod.switchsound")

	for i=1, 25 do
		handgunprojectileHandler.shells[i] = deepcopy(handgunprojectileHandler.defaultShell)
	end

	shootTimer = 0
	rocketTimer = 0
	recoilTimer = 0
	lightTimer = 0
	flashTimer = 0

	bangSprite = LoadSprite("img/bang.png")
	kaboomSprite = LoadSprite("img/kaboom.png")
	pewSound = LoadSound("MOD/snd/pew.ogg")
	powSound = LoadSound("MOD/snd/pow.ogg")
	hahaSound = LoadSound("MOD/snd/haha.ogg")
	kaSound = LoadSound("MOD/snd/ka.ogg")
	boomSound = LoadSound("MOD/snd/boom.ogg")
end

function deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else
        copy = orig
    end
    return copy
end

function GetAimPos()
	local ct = GetCameraTransform()
	local forwardPos = TransformToParentPoint(ct, Vec(0, 0, -100))
    local direction = VecSub(forwardPos, ct.pos)
    local distance = VecLength(direction)
	local direction = VecNormalize(direction)
	local hit, hitDistance = QueryRaycast(ct.pos, direction, distance)
	if hit then
		forwardPos = TransformToParentPoint(ct, Vec(0, 0, -hitDistance))
		distance = hitDistance
	end
	return forwardPos, hit, distance
end

function Shoot()
	if shootTimer > 0 or ammo == 0 then
		return
	end

	aimpos, hit, distance = GetAimPos()
	local direction = VecSub(aimpos, toolPos)

	handgunprojectileHandler.shells[handgunprojectileHandler.shellNum] = deepcopy(handgunprojectileHandler.defaultShell)
	loadedShell = handgunprojectileHandler.shells[handgunprojectileHandler.shellNum] 
	loadedShell.active = true
	loadedShell.pos = toolPos
	loadedShell.predictedBulletVelocity = VecScale(direction, velocity*(100/distance))
	loadedShell.rocket = false
	loadedShell.gravity = gravity

	handgunprojectileHandler.shellNum = (handgunprojectileHandler.shellNum%#handgunprojectileHandler.shells) +1
	SpawnParticle("fire", toolPos, Vec(0, 1.0+math.random(1,10)*0.1, 0), 0.5, 0.1)
	SpawnParticle("darksmoke", toolPos, Vec(0, 1.0+math.random(1,5)*0.1, 0), 0.4, 1.5)
	
	if not switchsound then
		PlaySound(pewSound, GetPlayerTransform().pos, 1.0, false)
	end
	
	if switchsound then
		PlaySound(powSound, GetPlayerTransform().pos, 1.0, false)
	end
	
	shootTimer = shotDelay
	recoilTimer = shotDelay
	lightTimer = shotDelay / 4
end

function ShootRocket()
	if rocketTimer > 0 or ammo == 0 then
		return
	end

	aimpos, hit, distance = GetAimPos()
	local direction = VecSub(aimpos, toolPos)

	handgunprojectileHandler.shells[handgunprojectileHandler.shellNum] = deepcopy(handgunprojectileHandler.defaultShell)
	loadedShell = handgunprojectileHandler.shells[handgunprojectileHandler.shellNum] 
	loadedShell.active = true
	loadedShell.pos = toolPos
	loadedShell.predictedBulletVelocity = VecScale(direction, rocketvelocity*(100/distance))
	loadedShell.rocket = true
	loadedShell.gravity = rocketgravity

	handgunprojectileHandler.shellNum = (handgunprojectileHandler.shellNum%#handgunprojectileHandler.shells) +1
	SpawnParticle("fire", toolPos, Vec(0, 1.0+math.random(1,10)*0.1, 0), 0.5, 0.1)
	SpawnParticle("darksmoke", toolPos, Vec(0, 1.0+math.random(1,5)*0.1, 0), 0.4, 1.5)
	PlaySound(kaSound, GetPlayerTransform().pos, 0.7, false)
	rocketTimer = shotDelay * 2
	recoilTimer = shotDelay * 1.5
	flashTimer = shotDelay / 2
end

function ProjectileOperations(projectile)
	projectile.predictedBulletVelocity = VecAdd(projectile.predictedBulletVelocity, (VecScale(projectile.gravity, GetTimeStep())))
	local point2 = VecAdd(projectile.pos, VecScale(projectile.predictedBulletVelocity, GetTimeStep()))
	local dir = VecNormalize(VecSub(point2, projectile.pos))
	local hit, dist = QueryRaycast(projectile.pos, dir, VecLength(VecSub(point2, projectile.pos)))
	
	if hit then
		hitPos = VecAdd(projectile.pos, VecScale(VecNormalize(VecSub(point2, projectile.pos)), dist))
		projectile.active = false
		if projectile.rocket then
			Explosion(hitPos, 3)
			PlaySound(boomSound, GetPlayerTransform().pos, 0.7, false)
		else
			MakeHole(hitPos, 2, 2, 2)
			SpawnParticle("smoke", hitPos, Vec(0, 1.0+math.random(1,10)*0.1, 0), damage, 1)
			if switchsound then
				--PlaySound(hahaSound, GetPlayerTransform().pos, 1.0, false)
			end
		end
	else
		DrawLine(projectile.pos, point2)
	end
    projectile.pos = point2
end

function tick(dt)
	if GetString("game.player.tool") == "handgun" and GetPlayerVehicle() == 0 then
		if InputPressed("lmb") then
			Shoot()
		end

		if InputPressed("rmb") then
			ShootRocket()
		end

		local b = GetToolBody()
		if b ~= 0 then
			local offset = Transform(Vec(0.1, 0.1, 0))
			SetToolTransform(offset)
			toolTrans = GetBodyTransform(b)
			toolPos = TransformToParentPoint(toolTrans, Vec(0.6, -0.45, -2.4))
			toolPos2 = TransformToParentPoint(toolTrans, Vec(0.3, -0.45, -3.4))

			if recoilTimer > 0 then
				local t = Transform()
				t.pos = Vec(0.1, 0.1, recoilTimer*3)
				t.rot = QuatEuler(recoilTimer*100, 0, 0)
				SetToolTransform(t)

				recoilTimer = recoilTimer - dt
			end

			if lightTimer > 0 then
				PointLight(toolPos, 1, 1, 1, 0.5)
				local rot = QuatLookAt(toolPos2, GetCameraTransform().pos)
				local transform = Transform(toolPos2, rot)
				DrawSprite(bangSprite, transform, 2, 2, 1.0, 1.0, 1.0, 0.9, true, false)
				lightTimer = lightTimer - dt
			end

			if flashTimer > 0 then
				PointLight(toolPos, 1, 1, 1, 0.5)
				local rot = QuatLookAt(GetCameraTransform().pos, toolPos2)
				local transform = Transform(toolPos2, rot)
				DrawSprite(kaboomSprite, transform, 2, 2, 1.0, 1.0, 1.0, 1, true, false)
				flashTimer = flashTimer - dt
			end
		end
	
		if shootTimer > 0 then shootTimer = shootTimer - dt end
		if rocketTimer > 0 then rocketTimer = rocketTimer - dt end
	end
	
	for key, shell in ipairs(handgunprojectileHandler.shells) do
		if shell.active then
			ProjectileOperations(shell)
		end
	end
end